/* Copyright (C) 2011 Ion Torrent Systems, Inc. All Rights Reserved */
/*
 * DPTreephaser.cpp
 *
 *  Created on: July 11, 2011 adapted from DPDephaser
 *      Author: ckoller
 */

/*
 * Minor modifications are made in the code for TopCoder marathon match usage.
 * They are marked by specific comments at the bottom of the file.
 *
 */

#include <cassert>
#include <cstdio>
#include <cmath>
#include <algorithm>

#include <iostream>
#include <string>
#include <vector>

using namespace std;

class BasecallerRead {
public:
  void SetDataAndKeyNormalize(const float *flowValues, int _numFlows, const int *keyVec, int numKeyFlows);

  void flowToString(const string &flowOrder, string &seq);

  void FitBaselineVector(int numSteps, int stepSize, int startStep = 0);
  void FitNormalizerVector(int numSteps, int stepSize, int startStep = 0);

  void Normalize(int fromFlow, int toFlow);


  int             numFlows;
  float           keyNormalizer;
  vector<float>   measurements;           // Measured, key-normalized flow intensities
  vector<char>    solution;               // HP-sequence determined by the solver. All entries are integer
  vector<float>   prediction;             // Predicted flow intensities corresponding to "solved" sequence
  vector<float>   normalizedMeasurements; // Measured flow intensities after more sophisticated normalization.

  float           miscNormalizer;

};

class TreephaserPath {
public:
  TreephaserPath(int numFlows) :
    state(numFlows), prediction(numFlows), solution(numFlows) {
    flow = 0;
    windowStart = 0;
    windowEnd = 0;
    pathMetric = 0.0;
    residualLeftOfWindow = 0.0;
    perFlowMetric = 0.0;
    dotCounter = 0;
    inUse = false;
  }

  bool inUse;

  // Phasing state of this path

  int flow; // in phase flow
  vector<float> state;
  int windowStart;
  int windowEnd;

  // Path metrics and related values

  float pathMetric;
  float residualLeftOfWindow; // Residual left of the window;
  float perFlowMetric;
  int dotCounter;

  // Input-output state of this path

  vector<float> prediction;
  vector<char>  solution;
};

class DPTreephaser {

public:
  DPTreephaser(const char *_flowOrder, int _numFlows, int _numPaths);
  ~DPTreephaser() {
  }

  void SetModelParameters(double cf, double ie, double dr);

  void NormalizeAndSolve5(BasecallerRead& read, int maxFlows); // Windowed adaptive normalization


  void Solve(BasecallerRead& read, int maxFlows, int restartFlows = 0);


  vector<double> HpScaleWeight; // For Normalization
  vector<float> oneMerHeight; // For QV metrics
  vector<float> penaltyMismatch; // For QV metrics: min penalty to earlier base (deletion before this base)
  vector<float> penaltyResidual; // For QV metrics: min penalty to later base (this base is insertion)

protected:

  void advanceState(TreephaserPath *child, const TreephaserPath *parent, int nuc, int toflow);

  int numFlows; // Number of flows
  int numPaths; // Maximum number of paths considered
  int usePaths;

  float Multiplier; // Multiplier used for normalization
  float minDistance; // Squared distance of solution to measurements
  float expThreshold; // Threshold for expanding paths
  float negMultiplier; // Extra weight on the negative residuals
  float dotValue; // percentage of expected Signal that constitutes a "dot"
  int maxHP; // maximum possible homopolymer length
  float minFraction; // Minimum fraction to be taken into account

  int maxPathDelay; // Paths that are delayed more are killed

  vector<TreephaserPath> path;

  vector<float> transitionBase[4], transitionFlow[4];
  vector<int> flowOrder;

};


DPTreephaser::DPTreephaser(const char *_flowOrder, int _numFlows, int _numPaths)
{
  numFlows = _numFlows;
  numPaths = _numPaths;

  // Set default values to options
  expThreshold = 0.2;
  minFraction = 1e-6;
  maxPathDelay = 40;
  negMultiplier = 2.0;
  dotValue = 0.3;
  maxHP = 11;
  usePaths = numPaths;

  float HpScale[5] = { 0, 1, 0.8, 0.64, 0.37 };
  HpScaleWeight.assign(HpScale, HpScale + 5);
  Multiplier = 1;

  // Initialize state storage
  for (int i = 0; i < 4; i++) {
    transitionBase[i].resize(numFlows);
    transitionFlow[i].resize(numFlows);
  }

  flowOrder.resize(numFlows);
  int iFlowShort = 0;
  for (int iFlow = 0; iFlow < numFlows; iFlow++) {
    switch (_flowOrder[iFlowShort]) {
    case 'T':
      flowOrder[iFlow] = 0;
      break;
    case 'A':
      flowOrder[iFlow] = 1;
      break;
    case 'C':
      flowOrder[iFlow] = 2;
      break;
    case 'G':
      flowOrder[iFlow] = 3;
      break;
    }
    iFlowShort++;
    if (_flowOrder[iFlowShort] == 0)
      iFlowShort = 0;
  }

  path.resize(numPaths, numFlows);
}

//-------------------------------------------------------------------------

void DPTreephaser::SetModelParameters(double cf, double ie, double dr)
{
  double distToBase[4] = { 0, 0, 0, 0 };
  for (int iFlow = 0; iFlow < numFlows; iFlow++) {
    distToBase[flowOrder[iFlow]] = 1;
    for (int nuc = 0; nuc < 4; nuc++) {
      transitionBase[nuc][iFlow] = distToBase[nuc] * (1 - dr) * (1 - ie);
      transitionFlow[nuc][iFlow] = (1 - distToBase[nuc]) + distToBase[nuc] * (1 - dr) * ie;
      distToBase[nuc] *= cf;
    }
  }
}

//-------------------------------------------------------------------------



void BasecallerRead::SetDataAndKeyNormalize(const float *flowValues, int _numFlows, const int *keyVec, int numKeyFlows)
{
  numFlows = _numFlows;
  measurements.resize(numFlows);
  normalizedMeasurements.resize(numFlows);
  solution.assign(numFlows, 0);
  prediction.assign(numFlows, 0);

  float onemerSum = 0.0;
  float onemerCount = 0.0;

  for (int iFlow = 0; iFlow < numKeyFlows; iFlow++) {
    if (keyVec[iFlow] == 1) {
      onemerSum += flowValues[iFlow];
      onemerCount += 1.0;
    }
  }

  keyNormalizer = 1;
  miscNormalizer = 1;
  for (int iFlow = 0; iFlow < numFlows; iFlow++)
    measurements[iFlow] = flowValues[iFlow];

  if ((onemerSum > 0) && (onemerCount > 0)) {
    keyNormalizer = onemerCount / onemerSum;

    for (int iFlow = 0; iFlow < numFlows; iFlow++)
      measurements[iFlow] *= keyNormalizer;
  }

  for (int iFlow = 0; iFlow < numFlows; iFlow++)
    normalizedMeasurements[iFlow] = measurements[iFlow];
}

void BasecallerRead::flowToString(const string &flowOrder, string &seq)
{
  seq = "";
  for(unsigned int iFlow=0; iFlow < solution.size(); iFlow++) {
    if(solution[iFlow] > 0) {
      char base = flowOrder[iFlow % flowOrder.length()];
      for(int iNuc = solution[iFlow]; iNuc > 0; iNuc--)
        seq += base;
    }
  }
}



// ----------------------------------------------------------------------
// New normalization strategy

float medianSet[111111];

void BasecallerRead::FitBaselineVector(int numSteps, int stepSize, int startStep)
{
  assert(numSteps > 0);

  float latestNormalizer = 0;
  int fromFlow = 0;
  int toFlow = stepSize;

  float* stepNormalizers = new float[numSteps];


  for (int iStep = startStep; iStep < numSteps; iStep++) {

    if (fromFlow >= numFlows)
      break;
    toFlow = std::min(toFlow, numFlows);

    if (toFlow < numFlows) {
      int medianCount = 0;
      for (int iFlow = fromFlow; iFlow < toFlow; iFlow++)
        if (prediction[iFlow] < 0.3)
          medianSet[medianCount++] = measurements[iFlow] - prediction[iFlow];

      if (medianCount > 5) {
        std::nth_element(medianSet, medianSet + medianCount/2, medianSet + medianCount);
        latestNormalizer = medianSet[medianCount / 2];
      }
    }

    stepNormalizers[iStep] = latestNormalizer;

    fromFlow = toFlow;
    toFlow += stepSize;
  }

  // Three modes:
  //  1 - before the middle of the first region: flat
  //  2 - in between regions: linear interpolation
  //  3 - beyond the middle of the last region: flat


  int halfStep = stepSize / 2;

  int iFlow = 0;
  if (startStep > 0)
    iFlow = halfStep + startStep * stepSize;

  for (; (iFlow < halfStep) && (iFlow < numFlows); iFlow++)
    normalizedMeasurements[iFlow] = measurements[iFlow] - stepNormalizers[0];

  for (int iStep = startStep + 1; iStep < numSteps; iStep++) {
    for (int pos = 0; (pos < stepSize) && (iFlow < numFlows); pos++, iFlow++)
      normalizedMeasurements[iFlow] = measurements[iFlow]
          - (stepNormalizers[iStep - 1] * (stepSize - pos) + stepNormalizers[iStep] * pos) / stepSize;
  }

  for (; iFlow < numFlows; iFlow++)
    normalizedMeasurements[iFlow] = measurements[iFlow] - stepNormalizers[numSteps - 1];

  delete[] stepNormalizers;
}

void BasecallerRead::FitNormalizerVector(int numSteps, int stepSize, int startStep)
{
  assert(numSteps > 0);

  float latestNormalizer = 1;
  int fromFlow = 0;
  int toFlow = stepSize;

  float* stepNormalizers = new float[numSteps];

  for (int iStep = startStep; iStep < numSteps; iStep++) {

    if (fromFlow >= numFlows)
      break;
    toFlow = std::min(toFlow, numFlows);

    if (toFlow < numFlows) {
      int medianCount = 0;
      for (int iFlow = fromFlow; iFlow < toFlow; iFlow++)
        if ((prediction[iFlow] > 0.5) && (normalizedMeasurements[iFlow] > 0))
          medianSet[medianCount++] = normalizedMeasurements[iFlow] / prediction[iFlow];

      if (medianCount > 5) {
        std::nth_element(medianSet, medianSet + medianCount/2, medianSet + medianCount);
        if (medianSet[medianCount / 2] > 0)
          latestNormalizer = medianSet[medianCount / 2];
      }
    }

    stepNormalizers[iStep] = latestNormalizer;

    fromFlow = toFlow;
    toFlow += stepSize;
  }

  // Three modes:
  //  1 - before the middle of the first region: flat
  //  2 - in between regions: linear interpolation
  //  3 - beyond the middle of the last region: flat

  float normalizer = 1;
  int halfStep = stepSize / 2;

  int iFlow = 0;
  if (startStep > 0)
    iFlow = halfStep + startStep * stepSize;

  for (; (iFlow < halfStep) && (iFlow < numFlows); iFlow++) {
    normalizer = stepNormalizers[0];
    normalizedMeasurements[iFlow] /= normalizer;
  }

  for (int iStep = startStep + 1; iStep < numSteps; iStep++) {
    for (int pos = 0; (pos < stepSize) && (iFlow < numFlows); pos++, iFlow++) {
      normalizer = stepNormalizers[iStep - 1] * (stepSize - pos) + stepNormalizers[iStep] * pos;
      normalizer /= stepSize;
      normalizedMeasurements[iFlow] /= normalizer;
    }
  }

  for (; iFlow < numFlows; iFlow++) {
    normalizer = stepNormalizers[numSteps - 1];
    normalizedMeasurements[iFlow] /= normalizer;
  }

  delete[] stepNormalizers;
}


// Sliding window adaptive normalization
void DPTreephaser::NormalizeAndSolve5(BasecallerRead& well, int maxFlows)
{
  int stepSize = 50;

  for (int iStep = 1; iStep < 100; iStep++) {

      int solveLength = (iStep+1) * stepSize;
      solveLength = std::min(solveLength, maxFlows);

      int restartLength = std::max(solveLength - 100, 0);
      int restartStep = 0; //std::max(iStep - 3, 0);

      Solve(well,solveLength, restartLength);
      well.FitBaselineVector(iStep, stepSize, restartStep);
      well.FitNormalizerVector(iStep, stepSize, restartStep);

      if (solveLength == maxFlows)
          break;
  }

  Solve(well,maxFlows);
}




void BasecallerRead::Normalize(int fromFlow, int toFlow)
{
  const float HpScaleWeight[5] = { 0, 1, 0.8, 0.64, 0.37 };

  double xy = 0;
  double yy = 0;

  for (int iFlow = fromFlow; ((iFlow < toFlow) && (iFlow < numFlows)); iFlow++) {
    if ((solution[iFlow] > 0) && (((int)solution[iFlow]) < 5)) {

      xy += HpScaleWeight[(int)solution[iFlow]] * measurements[iFlow] * prediction[iFlow];
//      xy += HpScaleWeight[(int)solution[iFlow]] * normalizedMeasurements[iFlow] * prediction[iFlow];
      yy += HpScaleWeight[(int)solution[iFlow]] * prediction[iFlow] * prediction[iFlow];
    }
  }

  double Divisor = 1;
  if ((yy > 0) && (xy > 0))
    Divisor = xy / yy;

  for (int iFlow = 0; iFlow < numFlows; iFlow++)
    normalizedMeasurements[iFlow] = measurements[iFlow] / Divisor;
//    normalizedMeasurements[iFlow] /= Divisor;

//  miscNormalizer *= Divisor;
  miscNormalizer = Divisor;
}


//-------------------------------------------------------------------------

void DPTreephaser::advanceState(TreephaserPath *child, const TreephaserPath *parent, int nuc, int toflow)
{
  if (child != parent) {  // The advance is not done in place

    // Advance flow
    child->flow = parent->flow;
    while ((child->flow < toflow) && (flowOrder[child->flow] != nuc))
      child->flow++;

    // Initialize window
    child->windowStart = parent->windowStart;
    child->windowEnd = parent->windowEnd;

    if (parent->flow == child->flow) {

      // This nuc simply prolongs current homopolymer, inherits state from parent
      for (int iFlow = child->windowStart; iFlow < child->windowEnd; iFlow++)
        child->state[iFlow] = parent->state[iFlow];
      return;
    }

    // This nuc begins a new homopolymer
    float alive = 0;
    for (int iFlow = parent->windowStart; iFlow < child->windowEnd; iFlow++) {

      // State progression according to phasing model
      if (iFlow < parent->windowEnd)
        alive += parent->state[iFlow];
      child->state[iFlow] = alive * transitionBase[nuc][iFlow];
      alive *= transitionFlow[nuc][iFlow];

      // Window maintenance
      if ((iFlow == child->windowStart) && (iFlow < (child->windowEnd - 1)) && (child->state[iFlow] < minFraction))
        child->windowStart++;

      if ((iFlow == (child->windowEnd - 1)) && (child->windowEnd < toflow) && (alive > minFraction))
        child->windowEnd++;
    }

  } else {    // The advance is done in place

    // Advance flow
    int oldFlow = child->flow;
    while ((child->flow < toflow) && (flowOrder[child->flow] != nuc))
      child->flow++;

    if (oldFlow == child->flow) // Same homopolymer, no changes in the state
      return;

    // This nuc begins a new homopolymer
    float alive = 0;
    int oldWindowStart = child->windowStart;
    int oldWindowEnd = child->windowEnd;
    for (int iFlow = oldWindowStart; iFlow < child->windowEnd; iFlow++) {

      // State progression according to phasing model
      if (iFlow < oldWindowEnd)
        alive += child->state[iFlow];
      child->state[iFlow] = alive * transitionBase[nuc][iFlow];
      alive *= transitionFlow[nuc][iFlow];

      // Window maintenance
      if ((iFlow == child->windowStart) && (iFlow < (child->windowEnd - 1)) && (child->state[iFlow] < minFraction))
        child->windowStart++;

      if ((iFlow == (child->windowEnd - 1)) && (child->windowEnd < toflow) && (alive > minFraction))
        child->windowEnd++;
    }
  }
}





// Solve - main tree search procedure that determines the base sequence.
// Another temporary version, uses external class for storing read data

void DPTreephaser::Solve(BasecallerRead& read, int maxFlows, int restartFlows)
{
  assert(maxFlows <= numFlows);

  // Initialize stack: just one root path
  for (int iPath = 1; iPath < numPaths; iPath++)
    path[iPath].inUse = false;

  path[0].flow = 0;
  path[0].state[0] = 1;
  path[0].windowStart = 0;
  path[0].windowEnd = 1;
  path[0].pathMetric = 0;
  path[0].perFlowMetric = 0;
  path[0].residualLeftOfWindow = 0;
  path[0].dotCounter = 0;
  path[0].prediction.assign(numFlows, 0);
  path[0].solution.assign(numFlows, 0);
  path[0].inUse = true;

  int spaceOnStack = usePaths - 1;
  minDistance = maxFlows;

  if (restartFlows > 0) {
    // The solver will not attempt to solve initial restartFlows
    // - Simulate restartFlows instead of solving
    // - If it turns out that solving was finished before restartFlows, simply exit.

    assert(read.solution.size() >= (unsigned int)restartFlows);
    int lastIncorporatingFlow = 0;

    for (int mainFlow = 0; (mainFlow < numFlows) && (mainFlow < restartFlows); mainFlow++) {

      path[0].solution[mainFlow] = read.solution[mainFlow];

      for (int iHP = 0; iHP < path[0].solution[mainFlow]; iHP++) {

        advanceState(&path[0], &path[0], flowOrder[mainFlow], maxFlows);

        for (int iFlow = path[0].windowStart; iFlow < path[0].windowEnd; iFlow++)
          path[0].prediction[iFlow] += path[0].state[iFlow];

        lastIncorporatingFlow = mainFlow;
      }
    }

    if (lastIncorporatingFlow < (restartFlows-10)) { // This read ended before restartFlows. No point resolving it.
      read.prediction.swap(path[0].prediction);
      return;
    }

    for (int iFlow = 0; iFlow < path[0].windowStart; iFlow++) {
      float residual = read.normalizedMeasurements[iFlow] - path[0].prediction[iFlow];
      path[0].residualLeftOfWindow += residual * residual;
    }
  }

  // Initializing variables
  read.solution.assign(numFlows, 0);
  read.prediction.assign(numFlows, 0);

  // Main loop to select / expand / delete paths
  while (1) {

    // ------------------------------------------
    // Step 1: Prune the content of the stack and make sure there are at least 4 empty slots

    // Remove paths that are more than 'maxPathDelay' behind the longest one
    if (spaceOnStack < (numPaths - 3)) {
      int maxFlow = 0;
      for (int iPath = 0; iPath < usePaths; iPath++)
        if (path[iPath].inUse)
          maxFlow = max(maxFlow, path[iPath].flow);

      if (maxFlow > maxPathDelay) {
        for (int iPath = 0; iPath < usePaths; iPath++) {
          if (path[iPath].inUse && (path[iPath].flow < (maxFlow - maxPathDelay))) {
            path[iPath].inUse = false;
            spaceOnStack++;
          }
        }
      }
    }

    // If necessary, remove paths with worst perFlowMetric
    while (spaceOnStack < 4) {
      // find maximum per flow metric
      float ExtremeVal = -0.1;
      int newIdx = usePaths;
      for (int iPath = 0; iPath < usePaths; iPath++) {
        if (path[iPath].inUse && (path[iPath].perFlowMetric > ExtremeVal)) {
          ExtremeVal = path[iPath].perFlowMetric;
          newIdx = iPath;
        }
      }

      // killing path with largest per flow metric
      if (!(newIdx < usePaths)) {
        printf("Failed assertion in Treephaser\n");
        for (int iPath = 0; iPath < usePaths; iPath++) {
          if (path[iPath].inUse)
            printf("Path %d, inUse = true, perFlowMetric = %f\n", iPath, path[iPath].perFlowMetric);
          else
            printf("Path %d, inUse = false, perFlowMetric = %f\n", iPath, path[iPath].perFlowMetric);
        }
        fflush(NULL);
      }

      assert (newIdx < usePaths);
      path[newIdx].inUse = false;
      spaceOnStack++;
    }

    // ------------------------------------------
    // Step 2: Select a path to expand or break if there is none

    TreephaserPath *parent = NULL;
    float ExtremeVal = 1000;
    for (int iPath = 0; iPath < usePaths; iPath++) {
      if (path[iPath].inUse && (path[iPath].pathMetric < ExtremeVal)) {
        ExtremeVal = path[iPath].pathMetric;
        parent = &(path[iPath]);
      }
    }
    if (!parent)
      break;

    // ------------------------------------------
    // Step 3: Construct four expanded paths and calculate feasibility metrics
    assert (spaceOnStack >= 4);

    TreephaserPath *children[4];

    for (int nuc = 0, iPath = 0; nuc < 4; iPath++)
      if (!path[iPath].inUse)
        children[nuc++] = &(path[iPath]);

    float penalty[4] = { 0, 0, 0, 0 };

    for (int nuc = 0; nuc < 4; nuc++) {

      TreephaserPath *child = children[nuc];

      advanceState(child, parent, nuc, maxFlows);

      // Apply easy termination rules

      if (child->flow >= maxFlows) {
        penalty[nuc] = 25; // Mark for deletion
        continue;
      }

      int currentHP = parent->solution[child->flow];
      if (currentHP == maxHP) {
        penalty[nuc] = 25; // Mark for deletion
        continue;
      }

      child->pathMetric = parent->residualLeftOfWindow;
      child->residualLeftOfWindow = parent->residualLeftOfWindow;

      float penaltyN = 0;
      float penalty1 = 0;

      for (int iFlow = parent->windowStart; (iFlow < child->windowEnd) ; iFlow++) {

        child->prediction[iFlow] = parent->prediction[iFlow] + child->state[iFlow];

        float residual = read.normalizedMeasurements[iFlow] - child->prediction[iFlow];
        float sqResidual = residual * residual;

        // Metric calculation
        if (iFlow < child->windowStart) {
          child->residualLeftOfWindow += sqResidual;
          child->pathMetric += sqResidual;
        } else if (residual <= 0)
          child->pathMetric += sqResidual;

        if (residual <= 0)
          penaltyN += sqResidual;
        else if (iFlow < child->flow)
          penalty1 += sqResidual;
      }


      penalty[nuc] = penalty1 + negMultiplier * penaltyN;
      penalty1 += penaltyN;

      child->perFlowMetric = (child->pathMetric + 0.5 * penalty1) / child->flow;

    } //looping over nucs


    // Find out which nuc has the least penalty (the greedy choice nuc)
    int bestNuc = 0;
    if (penalty[bestNuc] > penalty[1])
      bestNuc = 1;
    if (penalty[bestNuc] > penalty[2])
      bestNuc = 2;
    if (penalty[bestNuc] > penalty[3])
      bestNuc = 3;

    // ------------------------------------------
    // Step 4: Use calculated metrics to decide which paths are worth keeping

    for (int nuc = 0; nuc < 4; nuc++) {

      TreephaserPath *child = children[nuc];

      // Path termination rules

      if (penalty[nuc] >= 20)
        continue;

      if (child->pathMetric > minDistance)
        continue;

      // This is the only rule that depends on finding the "best nuc"
      if ((penalty[nuc] - penalty[bestNuc]) >= expThreshold)
        continue;

      float newSignal = (read.normalizedMeasurements[child->flow] - parent->prediction[child->flow]) / child->state[child->flow];
      child->dotCounter = (newSignal < dotValue) ? (parent->dotCounter + 1) : 0;
      if (child->dotCounter > 1)
        continue;

      //            if ((nuc != bestNuc) && (child->dotCounter > 0))
      //                continue;

      // Path survived termination rules and will be kept on stack
      child->inUse = true;
      spaceOnStack--;

      // Fill out the remaining portion of the prediction
      for (int iFlow = 0; iFlow < parent->windowStart; iFlow++)
        child->prediction[iFlow] = parent->prediction[iFlow];

      for (int iFlow = child->windowEnd; iFlow < maxFlows; iFlow++)
        child->prediction[iFlow] = 0;

      // Fill out the solution
      child->solution = parent->solution;
      child->solution[child->flow]++;
    }

    // ------------------------------------------
    // Step 5. Check if the selected path is in fact the best path so far

    // Computing sequence squared distance
    float SequenceDist = parent->residualLeftOfWindow;
    for (int iFlow = parent->windowStart; iFlow < maxFlows; iFlow++) {
      float residual = read.normalizedMeasurements[iFlow] - parent->prediction[iFlow];
      SequenceDist += residual * residual;
    }

    // Updating best path
    if (SequenceDist < minDistance) {
      read.prediction.swap(parent->prediction);
      read.solution.swap(parent->solution);
      minDistance = SequenceDist;
    }

    parent->inUse = false;
    spaceOnStack++;

  } // main decision loop
}

// Start of modifications for TopCoder marathon match usage.

vector<char> processt(float* flow, string order, float CF, float IE, int numFlows, int libNumKeyFlows, int* libKeyFlows) {
	DPTreephaser t(order.c_str(), numFlows, 8);
    t.SetModelParameters(CF, IE, 0);
    BasecallerRead s;
    s.SetDataAndKeyNormalize(flow, numFlows, libKeyFlows, libNumKeyFlows);
    t.NormalizeAndSolve5(s, numFlows);
    return s.solution;
}

class Treephaser {
public:
  vector <int> process(double CF, double IE, string flowOrder, int numFlows, int numKeyFlows, vector <int> keyVec, vector <double> flow) {
    float* cflow = new float[numFlows];
    for (int i = 0; i < numFlows; i++) {
      cflow[i] = flow[i];
    }

    vector<char> res = processt(cflow, flowOrder, CF, IE, numFlows, numKeyFlows, &(keyVec[0]));

    vector<int> cres(numFlows);
    for (int i = 0; i < numFlows; i++) {
      cres[i] = res[i];
    }

    return cres;
  }
};

// End of modifications for TopCoder marathon match usage.
